import dectate
import reg
import functools
import jsonpath_ng
import copy


def jsonpath_value(path, data, multi=False, default=None):
    if path == '$.':
        return data
    expr = jsonpath_ng.parse(path)
    res = expr.find(data)
    values = list([r.value for r in res])
    if multi:
        return values
    if res:
        return values[0]
    return default


class TransformerAction(dectate.Action):

    app_class_arg = True

    def __init__(self, name):
        self.name = name

    def identifier(self, app_class):
        return str([app_class, self.name])

    def perform(self, obj, app_class):
        def o(function, *args, **kwargs): return obj(*args, **kwargs)
        app_class._transform.register(name=self.name)(reg.methodify(o))


class Engine(dectate.App):

    register = dectate.directive(TransformerAction)

    def transform(self, name):
        return functools.partial(self._transform, name)

    def remap(self, rule, source, dest, **kwargs):
        rules = self._parse_rules(rule, **kwargs)
        dest = copy.deepcopy(dest)
        for p, f in rules:
            dest_pattern = jsonpath_ng.parse(p)
            # default = jsonpath_value(p, dest)
            # FIXME: get a sane default value
            default = None
            dest = dest_pattern.update(dest, f(source, dest, default))
        return dest

    def _return_value(self, v):
        return lambda s, d, dv: v

    def _source_path(self, path):
        def f(s, d, dv):
            return jsonpath_value(path, s, default=dv)
        return f

    def _apply_function(self, config,  **kwargs):
        def f(s, d, dv):
            params = {}
            for k, v in kwargs.items():
                params[k] = v
            for k, v in config.items():
                if isinstance(v, dict) and v.get('function', None):
                    v = self._apply_function(v, **kwargs)(s, d, dv)
                elif v.startswith('$.'):
                    v = jsonpath_value(v, s, default=dv)
                params[k] = v
            return self._transform(**params)
        return f

    def _parse_rules(self, rule, **kwargs):
        res = []
        for k, v in rule.items():
            if isinstance(v, dict) and v.get('function', None):
                res.append((k, self._apply_function(v, **kwargs)))
            elif isinstance(v, int) or isinstance(v, float):
                res.append((k, self._return_value(v)))
            elif v.startswith('$.'):
                res.append((k, self._source_path(v)))
            else:
                res.append((k, self._return_value(v)))
        return res

    @reg.dispatch_method(
        reg.match_key('name',
                      lambda self, function, *args, **kwargs: function))
    def _transform(self, function, *args, **kwargs):
        raise NotImplementedError('Transform function %s' % function)



from ... import operator as rop
from ...engine import Engine
import operator as op
from jsonpath_ng import parse as jsonpath_parse

METHOD = 'native'


@Engine.operator_compiler(method=METHOD, operator=rop.And)
def and_(engine, method, operator):
    def func(data):
        a = engine.compile_operator(method, operator.value[0])(data)
        for b in operator.value[1:]:
            a = op.__and__(a, engine.compile_operator(method, b)(data))
        return a
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Or)
def or_(engine, method, operator):
    def func(data):
        a = engine.compile_operator(method, operator.value[0])(data)
        for b in operator.value[1:]:
            a = op.__or__(a, engine.compile_operator(method, b)(data))
        return a
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Equal)
def eq(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return op.eq(sourceval, value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.NotEqual)
def ne(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return op.ne(sourceval, value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessEqualThan)
def le(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        return op.le(data[operator.field], value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterEqualThan)
def ge(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return op.ge(sourceval, value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessThan)
def lt(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return op.lt(sourceval, value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterThan)
def gt(engine, method, operator):
    def func(data):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(data)
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return op.gt(sourceval, value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.In)
def in_(engine, method, operator):
    def func(data):
        value = operator.value
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return sourceval in value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Get)
def get(engine, method, operator):
    return lambda data: data[operator.value]


@Engine.operator_compiler(method=METHOD, operator=rop.Like)
def like(engine, method, operator):
    def func(data):
        value = operator.value
        if operator.field.startswith('$'):
            match = jsonpath_parse(operator.field).find(data)
            if match:
                sourceval = match[0].value
        else:
            sourceval = data[operator.field]
        return value in sourceval
    return func



import shlex
import re
import boolean
import json
from .engine import OperatorNotAllowedError


class AND(boolean.AND):

    def json(self, allowed_operators=None):
        allowed_operators = allowed_operators or []
        values = []
        for a in self.args:
            j = a.json(allowed_operators)
            if allowed_operators and j['operator'].lower() not in allowed_operators:
                raise OperatorNotAllowedError(j['operator'])
            values.append(j)
        return {
            'operator': 'and',
            'value': values
        }


class OR(boolean.OR):

    def json(self, allowed_operators=None):
        allowed_operators = allowed_operators or []
        values = []
        for a in self.args:
            j = a.json(allowed_operators)
            if allowed_operators and j['operator'].lower() not in allowed_operators:
                raise OperatorNotAllowedError(j['operator'])
            values.append(j)
        return {
            'operator': 'or',
            'value': values
        }


float_pattern = re.compile(r'^\d+\.\d+$')
int_pattern = re.compile(r'^\d+$')

class FIELD(boolean.Symbol):

    def decode_symbol(self):
        for op in ['==', '<=', '>=', '!=', '<', '>', '=', ' in ']:
            if op in self.obj:
                ss = self.obj.split(op)
                if len(ss) != 2:
                    raise ValueError('Unable to decode "%s"' % self.obj)
                k, v = ss
                if op.strip() in ['in']:
                    ov = v.strip()
                    if ov.startswith('(') or ov.startswith('['):
                        ov = ov[1:]
                    if ov.endswith(')') or ov.endswith(']'):
                        ov = ov[:-1]
                    ov = ov.strip().split(',')
                    v = []
                    for vv in ov:
                        vv = vv.strip()
                        if float_pattern.match(vv):
                            vv = float(vv)
                        elif int_pattern.match(vv):
                            vv = int(vv)
                        v.append(vv)
                else:
                    v = v.strip()
                return k.strip(), op.strip(), v
        raise ValueError("Unable to decode symbol '%s'" % self.obj)

    def json(self, allowed_operators=None):
        allowed_operators = allowed_operators or []
        s = self.decode_symbol()
        k, o, v = s
        value = v
        if isinstance(value, str):
            if re.match(r'^[\d\.]+$', value):
                if value.find('.') == 1:
                    value = float(value)
                elif value.find('.') < 1:
                    value = int(value)
        if allowed_operators and o.lower() not in allowed_operators:
            raise OperatorNotAllowedError(o)
        return {
            'field': k,
            'operator': o,
            'value': value
        }


class BooleanAlgebra(boolean.BooleanAlgebra):

    def __init__(self):
        super(BooleanAlgebra, self).__init__(
            Symbol_class=FIELD, AND_class=AND, OR_class=OR)

    def tokenize(self, s):
        ops = {
            'or': boolean.TOKEN_OR,
            'and': boolean.TOKEN_AND,
            '(': boolean.TOKEN_LPAR,
            ')': boolean.TOKEN_RPAR
        }

        tokens = []
        splitted = shlex.split(s)

        for t in splitted:
            if t.lower() in ops.keys():
                tokens.append(t)
            else:
                if not tokens:
                    tokens.append(t)
                elif tokens[-1].lower() in ops.keys():
                    tokens.append(t)
                else:
                    tokens[-1] += ' %s' % t

        for col, tok in enumerate(tokens):
            if tok.lower() in ops:
                yield ops[tok.lower()], tok, col
            else:
                yield boolean.TOKEN_SYMBOL, tok, col


def parse_dsl(s, allowed_operators=None):
    allowed_operators = allowed_operators or []
    algebra = BooleanAlgebra()
    parsed = algebra.parse(s)
    res = parsed.json(allowed_operators)
    for i in res:
        if allowed_operators and (
                res['operator'].lower() not in allowed_operators):
            raise OperatorNotAllowedError(res['operator'])
    return res


class Field(object):

    def __init__(self, key):
        self.key = key

    def __eq__(self, value):
        return {
            'field': self.key,
            'operator': '==',
            'value': value
        }

    def __gt__(self, value):
        return {
            'field': self.key,
            'operator': '>',
            'value': value
        }

    def __lt__(self, value):
        return {
            'field': self.key,
            'operator': '<',
            'value': value
        }

    def __ge__(self, value):
        return {
            'field': self.key,
            'operator': '>=',
            'value': value
        }

    def __le__(self, value):
        return {
            'field': self.key,
            'operator': '<=',
            'value': value
        }

    def __ne__(self, value):
        return {
            'field': self.key,
            'operator': '!=',
            'value': value
        }

    def in_(self, values):
        return {
            'field': self.key,
            'operator': 'in',
            'value': values
        }


def or_(*args):
    return {
        'operator': 'or',
        'value': list(args)
    }


def and_(*args):
    return {
        'operator': 'and',
        'value': list(args)
    }


class FieldGetter(object):

    def __getitem__(self, key):
        return Field(key)


field = FieldGetter()



from ... import operator as rop
from ...engine import Engine
import sqlalchemy as sa

METHOD = 'sqlalchemy'


@Engine.operator_compiler(method=METHOD, operator=rop.And)
def and_(engine, method, operator):
    def func(model):
        c = []
        for v in operator.value:
            c.append(engine.compile_operator(method, v)(model))
        return sa.and_(*c)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Or)
def or_(engine, method, operator):
    def func(model):
        c = []
        for v in operator.value:
            c.append(engine.compile_operator(method, v)(model))
        return sa.or_(*c)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Equal)
def eq(engine, method, operator):
    def func(model):
        attr = getattr(model, operator.field)
        return attr == operator.value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.NotEqual)
def ne(engine, method, operator):
    def func(model):
        attr = getattr(model, operator.field)
        return attr != operator.value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessEqualThan)
def le(engine, method, operator):
    def func(model):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(model)
        attr = getattr(model, operator.field)
        return attr <= value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterEqualThan)
def ge(engine, method, operator):
    def func(model):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(model)
        attr = getattr(model, operator.field)
        return attr >= value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessThan)
def lt(engine, method, operator):
    def func(model):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(model)
        attr = getattr(model, operator.field)
        return attr < value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterThan)
def gt(engine, method, operator):
    def func(model):
        value = operator.value
        if isinstance(operator.value, rop.Operator):
            value = engine.compile_operator(method, operator.value)(model)
        attr = getattr(model, operator.field)
        return attr > value
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.In)
def in_(engine, method, operator):
    def func(model):
        attr = getattr(model, operator.field)
        return attr.in_(operator.value)
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Like)
def like(engine, method, operator):
    def func(model):
        attr = getattr(model, operator.field)
        return attr.ilike('%%%s%%' % operator.value)
    return func



#
from .engine import OperatorNotAllowedError
from .engine import NestedOperationNotAllowedError
from .compiler import native, sqlalchemy, elasticsearch
from . import operator
from .engine import Engine
from .compiler.native import operator as operator_compiler
from .dsl import field, or_, and_, parse_dsl
import dectate

dectate.commit(Engine)


def validate_condition(query, allowed_operators=None):
    engine = Engine()
    engine.validate_condition(query, allowed_operators)


def compile_condition(method, query, allowed_operators=None):
    engine = Engine()
    return engine.compile_condition(method, query, allowed_operators)




from ... import action as ra
from ... import operator as op
from ...engine import Engine

METHOD = 'native'


@Engine.rulechain_compiler(method=METHOD)
def _compile_rulechain(engine, method, rulechain):
    def apply(obj):
        for rule in rulechain:
            def condition(obj): return obj
            if rule.get('condition', None):
                condition = engine.compile_condition(
                    method, rule.get('condition', None))
            if condition(obj):
                actions = [engine.get_action(**act) for act in rule['actions']]
                for a in actions:
                    engine.compile_action(method, a)(obj)
                return obj
        return obj
    return apply


@Engine.action_compiler(method=METHOD, action=ra.Set)
def set(engine, method, action):
    def func(data):
        value = action.value
        if isinstance(value, op.Operator):
            value = engine.compile_operator(method, value)(data)
        data[action.field] = value
        return data
    return func



import dectate
import reg


class OperatorNotAllowedError(Exception):

    def __init__(self, message, *args, **kwargs):
        message = "Operator '%s' is not allowed" % message
        super(OperatorNotAllowedError, self).__init__(message, *args, **kwargs)


class NestedOperationNotAllowedError(Exception):

    def __init__(self, message, *args, **kwargs):
        message = "NestedOperation '%s' is not allowed" % message
        super(NestedOperationNotAllowedError, self).__init__(
            message, *args, **kwargs)


class OperatorAction(dectate.Action):

    app_class_arg = True

    def __init__(self, operator, type_):
        self.operator = operator
        self.type = type_

    def identifier(self, app_class):
        return str((app_class, self.operator, self.type))

    def perform(self, op, app_class):

        def operator(engine, operator, value, field=None):
            if field is not None:
                return op(operator, engine, field, value)
            return op(operator, engine, value)

        app_class.get_operator.register(operator,
                                        operator=self.operator,
                                        value=self.type)


class OperatorCompilerAction(dectate.Action):

    app_class_arg = True

    def __init__(self, method, operator):
        self.method = method
        self.operator = operator

    def identifier(self, app_class):
        return str((app_class, self.method, self.operator))

    def perform(self, obj, app_class):

        app_class.compile_operator.register(
            method=self.method,
            operator=self.operator)(obj)


class ActionAction(dectate.Action):

    app_class_arg = True

    def __init__(self, action):
        self.action = action

    def identifier(self, app_class):
        return str((app_class, self.action))

    def perform(self, obj, app_class):

        @app_class.get_action.register(action=self.action)
        def action(engine, action, parameter):
            return obj(self.action, engine, **parameter)


class ActionCompilerAction(dectate.Action):

    app_class_arg = True

    def __init__(self, method, action):
        self.method = method
        self.action = action

    def identifier(self, app_class):
        return str((app_class, self.method, self.action))

    def perform(self, obj, app_class):
        app_class.compile_action.register(
            method=self.method, action=self.action)(obj)


class RuleChainCompilerAction(dectate.Action):

    app_class_arg = True

    def __init__(self, method):
        self.method = method

    def identifier(self, app_class):
        return str((app_class, self.method))

    def perform(self, obj, app_class):

        app_class.compile_rulechain.register(
            method=self.method)(obj)


class Engine(dectate.App):

    _operator = dectate.directive(OperatorAction)

    rulechain_compiler = dectate.directive(RuleChainCompilerAction)
    action = dectate.directive(ActionAction)

    @classmethod
    def operator(klass, operator, types):
        def func(op):
            for t in types + [type(None)]:
                klass._operator(operator, t)(op)
            return op
        return func

    operator_compiler = dectate.directive(OperatorCompilerAction)

    action_compiler = dectate.directive(ActionCompilerAction)

    @reg.dispatch_method(
        reg.match_key('operator',
                      lambda self, operator, value, field: operator),
        reg.match_instance('value'))
    def get_operator(self, operator, value, field=None):
        raise NotImplementedError

    @reg.dispatch_method(
        reg.match_key('method',
                      lambda self, method, operator: method),
        reg.match_instance('operator'))
    def compile_operator(self, method, operator):
        raise NotImplementedError((method, operator))

    @reg.dispatch_method(
        reg.match_key('action', lambda self, action, parameter: action))
    def get_action(self, action, parameter):
        raise NotImplementedError((action, parameter))

    @reg.dispatch_method(
        reg.match_key('method', lambda self, method, action: method),
        reg.match_instance('action'))
    def compile_action(self, method, action):
        raise NotImplementedError((method, action))

    def parse_condition(self, config):
        param = {
            'operator': config['operator'],
            'value': config['value']
        }
        if config.get('field', None):
            param['field'] = config['field']

        parsed = self.get_operator(**param)
        return parsed

    def parse_action(self, config):
        return self.get_action(**config)

    def compile_condition(self, method, query, allowed_operators=None,
                          nestable_operators=None):
        self.validate_condition(query, allowed_operators, nestable_operators)
        q = self.parse_condition(query)
        return self.compile_operator(method, q)

    @reg.dispatch_method(
        reg.match_key('method', lambda self, method, rulechain: method))
    def compile_rulechain(self, method, rulechain):
        raise NotImplementedError

    def validate_condition(self, query, allowed_operators=None,
                           nestable_operators=None):
        parsed = self.parse_condition(query)
        self._validate_condition(parsed, allowed_operators, nestable_operators)

    def _validate_condition(self, query, allowed_operators=None,
                            nestable_operators=None):
        from .operator import Operator

        if (allowed_operators is not None and
                query.operator not in allowed_operators):
            raise OperatorNotAllowedError(query.operator)

        if isinstance(query.value, list) or isinstance(query.value, tuple):
            for v in query.value:
                if isinstance(v, Operator):
                    if (nestable_operators is not None and
                            query.operator in nestable_operators and
                            v.operator in nestable_operators):
                        raise NestedOperationNotAllowedError(query)
                    self._validate_condition(v, allowed_operators)



import reg
import operator as op
from .engine import Engine


class Operator(object):

    def __init__(self, operator, engine, value):
        self.operator = operator
        vals = []
        for v in (value or []):
            op = v['operator']
            value = v['value']
            field = v.get('field', None)
            o = engine.get_operator(op, value, field)
            vals.append(o)
        self.value = vals


@Engine.operator(operator='and', types=[list, tuple])
class And(Operator):
    pass


@Engine.operator(operator='or', types=[list, tuple])
class Or(Operator):
    pass


class FieldOperator(Operator):

    def __init__(self, operator, engine, field, value):
        self.operator = operator
        self.field = field
        if isinstance(value, dict) and value.get('operator', None):
            value = engine.get_operator(**value)
        self.value = value


@Engine.operator(operator='==', types=[dict, str, float, int])
class Equal(FieldOperator):
    pass


@Engine.operator(operator='!=', types=[dict, str, float, int])
class NotEqual(FieldOperator):
    pass


@Engine.operator(operator='<=', types=[dict, str, float, int])
class LessEqualThan(FieldOperator):
    pass


@Engine.operator(operator='>=', types=[dict, str, float, int])
class GreaterEqualThan(FieldOperator):
    pass


@Engine.operator(operator='<', types=[dict, str, float, int])
class LessThan(FieldOperator):
    pass


@Engine.operator(operator='>', types=[dict, str, float, int])
class GreaterThan(FieldOperator):
    pass


@Engine.operator(operator='in', types=[list, tuple])
class In(FieldOperator):
    pass


class FieldGetter(Operator):

    def __init__(self, operator, engine, value):
        self.operator = operator
        self.value = value


@Engine.operator(operator='get', types=[str])
class Get(FieldGetter):
    pass


@Engine.operator(operator='~', types=[str])
class Like(FieldOperator):
    pass



from ... import operator as rop
from ...engine import Engine
import operator as op

METHOD = 'elasticsearch'


@Engine.operator_compiler(method=METHOD, operator=rop.And)
def and_(engine, method, operator):
    def func():
        values = []
        for value in operator.value:
            vv = engine.compile_operator(method, value)()
            values += vv['bool']['must']
        return {
            'bool': {
                'must': values
            }
        }
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Or)
def or_(engine, method, operator):
    def func():
        values = []
        for value in operator.value:
            vv = engine.compile_operator(method, value)()
            values += vv['bool']['must']
        return {
            'bool': {
                'should': values,
                'minimum_should_match': 1
            }
        }
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Equal)
def eq(engine, method, operator):
    def func():
        return {'bool': {'must': [{'term': {operator.field: operator.value}}]}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterEqualThan)
def ge(engine, method, operator):
    def func():
        return {'bool': {'must': [
            {'range': {operator.field: {'gte': operator.value}}}]}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessEqualThan)
def le(engine, method, operator):
    def func():
        return {'bool': {'must': [
            {'range': {operator.field: {'lte': operator.value}}}]}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.GreaterThan)
def gt(engine, method, operator):
    def func():
        return {'bool': {'must': [
            {'range': {operator.field: {'gt': operator.value}}}]}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.LessThan)
def lt(engine, method, operator):
    def func():
        return {'bool': {'must': [
            {'range': {operator.field: {'lt': operator.value}}}]}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.In)
def in_(engine, method, operator):
    def func():
        return {'match': {operator.field: {'query': ' '.join(operator.value)}}}
    return func


@Engine.operator_compiler(method=METHOD, operator=rop.Like)
def like(engine, method, operator):
    def func():
        return {'match': {operator.field: operator.value}}
    return func



import reg
from .engine import Engine
from . import operator as op


class Action(object):
    pass


@Engine.action(action='set')
class Set(Action):

    @property
    def identifier(self):
        return ':'.join([self.action, self.field])

    def __init__(self, action, engine, field, value):
        self.action = action
        self.field = field
        if isinstance(value, dict):
            value = engine.get_operator(**value)
        self.value = value



from setuptools import setup, find_packages
import sys
import os

version = '0.1.2'

setup(name='rulez',
      version=version,
      description="Simple business rules engine configurable using YAML/JSON",
      long_description="""\
""",
      # Get strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers

      classifiers=[],

      keywords='business-rules rules reg',
      author='Izhar Firdaus',
      author_email='izhar@abyres.net',
      url='http://github.com/abyres/',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'reg',
          'dectate',
          'boolean.py',
          'jsonpath_ng',
          'sqlalchemy'
          # -*- Extra requirements: -*-
      ],
      extras_require={
          'test': [
              'mirakuru',
              'elasticsearch>=5.0.0,<6.0.0',
          ]
      },
      entry_points="""
      # -*- Entry points: -*-
      """,
      )




from ... import action as ra
from ... import operator as op
from ...engine import Engine
from sqlalchemy.sql import expression as sqlexp
import sqlalchemy as sa
METHOD = 'sqlalchemy'


@Engine.rulechain_compiler(method=METHOD)
def _compile_rulechain(engine, method, rulechain):
    def apply(model):
        conditions = {}
        defaults = {}
        for rule in rulechain:
            if rule.get('condition', None):
                condition = engine.compile_condition(
                    method, rule.get('condition', None))
                actions = [engine.get_action(**act) for act in rule['actions']]
                for a in actions:
                    conditions.setdefault(a.identifier, [])
                    conditions[a.identifier].append(
                        (condition(model),
                         engine.compile_action(method, a)(model)))
            else:
                actions = [engine.get_action(**act) for act in rule['actions']]
                for a in actions:
                    defaults[a.identifier] = engine.compile_action(
                        method, a)(model)
        result = []
        for k, v in conditions.items():
            d = defaults.get(k, None)
            result.append(sqlexp.case(v, else_=d).label(k))
        return result
    return apply


@Engine.action_compiler(method=METHOD, action=ra.Set)
def set(engine, method, action):
    def func(model):
        value = action.value
        if isinstance(value, op.Operator):
            value = engine.compile_operator(method, value)(model)
        return value
    return func
